import datetime
import uuid
from typing import Any

import pytest
from pytest_mock import MockerFixture

from ffun.application.resources import Resource as AppResource
from ffun.application.user_settings import UserSetting
from ffun.feeds_links import domain as fl_domain
from ffun.openai import errors
from ffun.openai.entities import APIKeyUsage, KeyStatus, UserKeyInfo
from ffun.openai.keys_rotator import (
    _api_key_is_working,
    _choose_user,
    _filter_out_users_for_whome_entry_is_too_old,
    _filter_out_users_with_overused_keys,
    _filter_out_users_with_wrong_keys,
    _filter_out_users_without_keys,
    _filters,
    _find_best_user_with_key,
    _get_candidates,
    _get_user_key_infos,
    _use_key,
    _users_for_feed,
    api_key_for_feed_entry,
)
from ffun.openai.keys_statuses import Statuses, statuses
from ffun.resources import domain as r_domain
from ffun.resources import entities as r_entities
from ffun.user_settings import domain as us_domain


class TestAPIKeyIsWorking:
    @pytest.mark.asyncio
    async def test_is_working(self, openai_key: str) -> None:
        statuses.set(openai_key, KeyStatus.works)
        assert await _api_key_is_working(openai_key)

    @pytest.mark.parametrize(
        "status", [status for status in KeyStatus if status not in [KeyStatus.works, KeyStatus.unknown]]
    )
    @pytest.mark.asyncio
    async def test_guaranted_broken(self, status: KeyStatus, openai_key: str) -> None:
        statuses.set(openai_key, status)
        assert not await _api_key_is_working(openai_key)

    @pytest.mark.parametrize("status, expected_result", [(KeyStatus.works, True), (KeyStatus.broken, False)])
    @pytest.mark.asyncio
    async def test_unknown(
        self, status: KeyStatus, expected_result: bool, statuses: Statuses, mocker: MockerFixture, openai_key: str
    ) -> None:
        assert statuses.get(openai_key) == KeyStatus.unknown

        mocker.patch("ffun.openai.client.check_api_key", return_value=status)

        assert await _api_key_is_working(openai_key) == expected_result


class TestFilterOutUsersWithWrongKeys:
    @pytest.mark.asyncio
    async def test_empty_list(self) -> None:
        assert await _filter_out_users_with_wrong_keys([]) == []

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        assert five_user_key_infos[1].api_key
        assert five_user_key_infos[3].api_key

        statuses.set(five_user_key_infos[1].api_key, KeyStatus.broken)
        statuses.set(five_user_key_infos[3].api_key, KeyStatus.quota)

        infos = await _filter_out_users_with_wrong_keys(five_user_key_infos)

        assert infos == [five_user_key_infos[i] for i in [0, 2, 4]]


class TestFilterOutUsersWithoutKeys:
    @pytest.mark.asyncio
    async def test_empty_list(self) -> None:
        assert await _filter_out_users_without_keys([]) == []

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        five_user_key_infos[1].api_key = None
        five_user_key_infos[3].api_key = None

        infos = await _filter_out_users_without_keys(five_user_key_infos)

        assert infos == [five_user_key_infos[i] for i in [0, 2, 4]]


class TestFilterOutUsersForWhomeEntryIsTooOld:
    @pytest.mark.asyncio
    async def test_empty_list(self) -> None:
        assert await _filter_out_users_for_whome_entry_is_too_old([], datetime.timedelta(days=1)) == []

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        for info, days in zip(five_user_key_infos, [5, 2, 3, 1, 4]):
            info.process_entries_not_older_than = datetime.timedelta(days=days)

        infos = await _filter_out_users_for_whome_entry_is_too_old(five_user_key_infos, datetime.timedelta(days=3))

        assert infos == [five_user_key_infos[i] for i in [0, 2, 4]]


class TestFilterOutUsersWithOverusedKeys:
    @pytest.mark.asyncio
    async def test_empty_list(self) -> None:
        assert await _filter_out_users_with_overused_keys([], 100) == []

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        for info, max_tokens_in_month in zip(five_user_key_infos, [201, 100, 300, 200, 500]):
            info.tokens_used = 50
            info.max_tokens_in_month = max_tokens_in_month

        infos = await _filter_out_users_with_overused_keys(five_user_key_infos, 150)

        assert infos == [five_user_key_infos[i] for i in [0, 2, 4]]


class TestUsersForFeed:
    @pytest.mark.asyncio
    async def test_no_users(self) -> None:
        assert await _users_for_feed(feed_id=uuid.uuid4()) == set()

    @pytest.mark.asyncio
    async def test_linked_users(self, five_internal_user_ids: list[uuid.UUID], saved_feed_id: uuid.UUID) -> None:
        indexes = [1, 3]

        for i in indexes:
            await fl_domain.add_link(five_internal_user_ids[i], saved_feed_id)

        assert await _users_for_feed(feed_id=saved_feed_id) == {five_internal_user_ids[i] for i in indexes}

    @pytest.mark.asyncio
    async def test_collections(
        self, five_internal_user_ids: list[uuid.UUID], saved_collection_feed_id: uuid.UUID
    ) -> None:
        indexes = [1, 3]

        await us_domain.remove_setting_for_all_users(UserSetting.openai_allow_use_key_for_collections)

        for i in indexes:
            await us_domain.save_setting(
                user_id=five_internal_user_ids[i], kind=UserSetting.openai_allow_use_key_for_collections, value=True
            )

        assert await _users_for_feed(feed_id=saved_collection_feed_id) == {five_internal_user_ids[i] for i in indexes}

    @pytest.mark.asyncio
    async def test_merging_users(
        self, five_internal_user_ids: list[uuid.UUID], saved_collection_feed_id: uuid.UUID
    ) -> None:
        await us_domain.remove_setting_for_all_users(UserSetting.openai_allow_use_key_for_collections)

        await fl_domain.add_link(five_internal_user_ids[1], saved_collection_feed_id)

        await us_domain.save_setting(
            user_id=five_internal_user_ids[3], kind=UserSetting.openai_allow_use_key_for_collections, value=True
        )

        assert await _users_for_feed(feed_id=saved_collection_feed_id) == {five_internal_user_ids[i] for i in [1, 3]}


class TestChooseUser:
    @pytest.mark.asyncio
    async def test_no_users(self) -> None:
        interval_started_at = r_domain.month_interval_start()

        with pytest.raises(errors.NoKeyFoundForFeed):
            assert await _choose_user(infos=[], reserved_tokens=0, interval_started_at=interval_started_at)

    @pytest.mark.asyncio
    async def test_no_users_with_resources(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        interval_started_at = r_domain.month_interval_start()

        with pytest.raises(errors.NoKeyFoundForFeed):
            assert await _choose_user(
                infos=five_user_key_infos,
                reserved_tokens=max(info.max_tokens_in_month for info in five_user_key_infos) + 1,
                interval_started_at=interval_started_at,
            )

    @pytest.mark.asyncio
    async def test_all_working(self, five_user_key_infos: list[UserKeyInfo]) -> None:
        interval_started_at = r_domain.month_interval_start()

        max_tokens = max(info.max_tokens_in_month for info in five_user_key_infos) + 1

        five_user_key_infos[2].max_tokens_in_month = max_tokens + five_user_key_infos[2].tokens_used

        info = await _choose_user(
            infos=five_user_key_infos, reserved_tokens=max_tokens, interval_started_at=interval_started_at
        )

        assert info == five_user_key_infos[2]


class TestUseKey:
    @pytest.mark.asyncio
    async def test_success(self, internal_user_id: uuid.UUID, openai_key: str) -> None:
        interval_started_at = r_domain.month_interval_start()

        reserved_tokens = 567

        await r_domain.try_to_reserve(
            user_id=internal_user_id,
            kind=AppResource.openai_tokens,
            interval_started_at=interval_started_at,
            amount=reserved_tokens,
            limit=1000,
        )

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.openai_tokens, interval_started_at=interval_started_at
        )

        used_tokens = 132

        async with _use_key(
            user_id=internal_user_id,
            api_key=openai_key,
            reserved_tokens=reserved_tokens,
            interval_started_at=interval_started_at,
        ) as key_usage:
            assert key_usage == APIKeyUsage(user_id=internal_user_id, api_key=openai_key, used_tokens=None)
            key_usage.used_tokens = used_tokens

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.openai_tokens, interval_started_at=interval_started_at
        )

        assert resources == {
            internal_user_id: r_entities.Resource(
                user_id=internal_user_id,
                kind=AppResource.openai_tokens,
                interval_started_at=interval_started_at,
                used=used_tokens,
                reserved=0,
            )
        }

    @pytest.mark.asyncio
    async def test_error(self, internal_user_id: uuid.UUID, openai_key: str) -> None:
        interval_started_at = r_domain.month_interval_start()

        reserved_tokens = 567

        await r_domain.try_to_reserve(
            user_id=internal_user_id,
            kind=AppResource.openai_tokens,
            interval_started_at=interval_started_at,
            amount=reserved_tokens,
            limit=1000,
        )

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.openai_tokens, interval_started_at=interval_started_at
        )

        class FakeError(Exception):
            pass

        with pytest.raises(FakeError):
            async with _use_key(
                user_id=internal_user_id,
                api_key=openai_key,
                reserved_tokens=reserved_tokens,
                interval_started_at=interval_started_at,
            ):
                raise FakeError()

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.openai_tokens, interval_started_at=interval_started_at
        )

        assert resources == {
            internal_user_id: r_entities.Resource(
                user_id=internal_user_id,
                kind=AppResource.openai_tokens,
                interval_started_at=interval_started_at,
                used=reserved_tokens,
                reserved=0,
            )
        }


class TestGetUserKeyInfos:
    @pytest.mark.asyncio
    async def test_no_users(self) -> None:
        interval_started_at = r_domain.month_interval_start()

        assert await _get_user_key_infos(user_ids=[], interval_started_at=interval_started_at) == []

    @pytest.mark.asyncio
    async def test_works(self, five_internal_user_ids: list[uuid.UUID]) -> None:
        from ffun.application.resources import Resource as AppResource
        from ffun.application.user_settings import UserSetting

        interval_started_at = r_domain.month_interval_start()

        keys = [uuid.uuid4().hex for _ in range(5)]
        max_tokens_in_month = [(i + 1) * 1000 for i in range(5)]
        days = list(range(5))
        used_tokens = [i * 100 for i in range(5)]
        reserved_tokens = [i * 10 for i in range(5)]

        for i, user_id in enumerate(five_internal_user_ids):
            await us_domain.save_setting(user_id=user_id, kind=UserSetting.openai_api_key, value=keys[i])

            await us_domain.save_setting(
                user_id=user_id, kind=UserSetting.openai_max_tokens_in_month, value=max_tokens_in_month[i]
            )

            await us_domain.save_setting(
                user_id=user_id, kind=UserSetting.openai_process_entries_not_older_than, value=days[i]
            )

            await r_domain.try_to_reserve(
                user_id=user_id,
                kind=AppResource.openai_tokens,
                interval_started_at=interval_started_at,
                amount=reserved_tokens[i],
                limit=max_tokens_in_month[i],
            )

            await r_domain.convert_reserved_to_used(
                user_id=user_id,
                kind=AppResource.openai_tokens,
                interval_started_at=interval_started_at,
                used=used_tokens[i],
                reserved=0,
            )

        infos = await _get_user_key_infos(five_internal_user_ids, interval_started_at)

        assert infos == [
            UserKeyInfo(
                user_id=five_internal_user_ids[i],
                api_key=keys[i],
                max_tokens_in_month=max_tokens_in_month[i],
                process_entries_not_older_than=datetime.timedelta(days=days[i]),
                tokens_used=used_tokens[i] + reserved_tokens[i],
            )
            for i in range(5)
        ]


class TestGetCandidates:
    @pytest.mark.asyncio
    async def test_no_users(self, saved_feed_id: uuid.UUID) -> None:
        interval_started_at = r_domain.month_interval_start()

        assert (
            await _get_candidates(
                feed_id=saved_feed_id,
                interval_started_at=interval_started_at,
                entry_age=datetime.timedelta(days=1),
                reserved_tokens=100,
            )
            == []
        )

    @pytest.mark.asyncio
    async def test_default_filters(self) -> None:
        assert _filters == (
            _filter_out_users_without_keys,
            _filter_out_users_for_whome_entry_is_too_old,
            _filter_out_users_with_wrong_keys,
            _filter_out_users_with_overused_keys,
        )

    @pytest.mark.asyncio
    async def test_filters_used(self, saved_feed_id: uuid.UUID, five_internal_user_ids: list[uuid.UUID]) -> None:
        for user_id in five_internal_user_ids:
            await fl_domain.add_link(user_id, saved_feed_id)

        interval_started_at = r_domain.month_interval_start()

        filter_1_users = [five_internal_user_ids[0], five_internal_user_ids[2], five_internal_user_ids[4]]
        filter_2_users = five_internal_user_ids
        filter_3_users = five_internal_user_ids[1:]

        def create_filter(filter_users: list[uuid.UUID]) -> Any:
            async def _filter(infos: list[UserKeyInfo], **kwargs: Any) -> list[UserKeyInfo]:
                return [info for info in infos if info.user_id in filter_users]

            return _filter

        infos = await _get_candidates(
            feed_id=saved_feed_id,
            interval_started_at=interval_started_at,
            entry_age=datetime.timedelta(days=1),
            reserved_tokens=100,
            filters=(create_filter(filter_1_users), create_filter(filter_2_users), create_filter(filter_3_users)),
        )

        assert {info.user_id for info in infos} == {five_internal_user_ids[2], five_internal_user_ids[4]}

    @pytest.mark.asyncio
    async def test_all_users_excluded(self, saved_feed_id: uuid.UUID, five_internal_user_ids: list[uuid.UUID]) -> None:
        for user_id in five_internal_user_ids:
            await fl_domain.add_link(user_id, saved_feed_id)

        interval_started_at = r_domain.month_interval_start()

        filter_1_users = [five_internal_user_ids[0], five_internal_user_ids[2], five_internal_user_ids[4]]

        def create_filter(filter_users: list[uuid.UUID]) -> Any:
            async def _filter(infos: list[UserKeyInfo], **kwargs: Any) -> list[UserKeyInfo]:
                return [info for info in infos if info.user_id in filter_users]

            return _filter

        async def _filter_3(infos: list[UserKeyInfo], **kwargs: Any) -> list[UserKeyInfo]:
            raise Exception("Should not be called")

        infos = await _get_candidates(
            feed_id=saved_feed_id,
            interval_started_at=interval_started_at,
            entry_age=datetime.timedelta(days=1),
            reserved_tokens=100,
            filters=(create_filter(filter_1_users), create_filter([]), _filter_3),
        )

        assert infos == []


class TestFindBestUserWithKey:
    @pytest.mark.asyncio
    async def test_no_users(self, saved_feed_id: uuid.UUID) -> None:
        with pytest.raises(errors.NoKeyFoundForFeed):
            await _find_best_user_with_key(
                feed_id=saved_feed_id,
                entry_age=datetime.timedelta(days=1),
                interval_started_at=r_domain.month_interval_start(),
                reserved_tokens=100,
            )

    @pytest.mark.asyncio
    async def test_works(self, saved_feed_id: uuid.UUID, five_user_key_infos: list[UserKeyInfo]) -> None:
        for info in five_user_key_infos:
            await fl_domain.add_link(info.user_id, saved_feed_id)

        chosen_users: set[uuid.UUID] = set()

        interval_started_at = r_domain.month_interval_start()

        used_tokens = 7

        for _ in range(len(five_user_key_infos)):
            info = await _find_best_user_with_key(
                feed_id=saved_feed_id,
                entry_age=datetime.timedelta(days=0),
                interval_started_at=interval_started_at,
                reserved_tokens=used_tokens,
            )

            assert info.tokens_used == five_user_key_infos[0].tokens_used

            chosen_users.add(info.user_id)

            await r_domain.convert_reserved_to_used(
                user_id=info.user_id,
                kind=AppResource.openai_tokens,
                interval_started_at=interval_started_at,
                used=used_tokens,
                reserved=used_tokens,
            )

        assert chosen_users == {info.user_id for info in five_user_key_infos}

        # next user will have more used tokens
        info = await _find_best_user_with_key(
            feed_id=saved_feed_id,
            entry_age=datetime.timedelta(days=0),
            interval_started_at=interval_started_at,
            reserved_tokens=used_tokens,
        )

        assert info.tokens_used == five_user_key_infos[0].tokens_used + used_tokens


class TestApiKeyForFeedEntry:
    @pytest.mark.asyncio
    async def test_no_users(self, saved_feed_id: uuid.UUID) -> None:
        with pytest.raises(errors.NoKeyFoundForFeed):
            async with api_key_for_feed_entry(
                feed_id=saved_feed_id, entry_age=datetime.timedelta(days=0), reserved_tokens=1
            ):
                pass

    @pytest.mark.asyncio
    async def test_works(self, saved_feed_id: uuid.UUID, five_user_key_infos: list[UserKeyInfo]) -> None:
        for info in five_user_key_infos:
            await fl_domain.add_link(info.user_id, saved_feed_id)

        chosen_users: set[uuid.UUID] = set()

        interval_started_at = r_domain.month_interval_start()

        used_tokens = 7

        for _ in range(len(five_user_key_infos)):
            async with api_key_for_feed_entry(
                feed_id=saved_feed_id, entry_age=datetime.timedelta(days=0), reserved_tokens=13
            ) as key_usage:
                key_usage.used_tokens = used_tokens
                chosen_users.add(key_usage.user_id)

        assert chosen_users == {info.user_id for info in five_user_key_infos}

        # next user will have more used tokens
        info = await _find_best_user_with_key(
            feed_id=saved_feed_id,
            entry_age=datetime.timedelta(days=0),
            interval_started_at=interval_started_at,
            reserved_tokens=1,
        )

        assert info.tokens_used == five_user_key_infos[0].tokens_used + used_tokens



from ffun.core.register import Entity, Register
from ffun.user_settings.types import Type


class Value(Entity):
    def __init__(
        self, key: int, name: str, type: Type, default: str | int | bool | None, description: str | None
    ) -> None:
        super().__init__(key=key)
        self.name = name
        self.type = type
        self.default = default
        self.description = description


SettingsRegister = Register[Value]


user_settings: SettingsRegister = Register()



import datetime
import uuid

from ffun.core import utils
from ffun.resources import operations
from ffun.resources.entities import Resource

load_resources = operations.load_resources
try_to_reserve = operations.try_to_reserve
convert_reserved_to_used = operations.convert_reserved_to_used
load_resource_history = operations.load_resource_history


def month_interval_start(now: datetime.datetime | None = None) -> datetime.datetime:
    if now is None:
        now = utils.now()

    return now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)


async def load_resource(user_id: uuid.UUID, kind: int, interval_started_at: datetime.datetime) -> Resource:
    resources = await load_resources([user_id], kind, interval_started_at)
    return resources[user_id]



import pkg_resources

from ffun.application.settings import settings


def user_agent() -> str:
    name = settings.app_name.replace(" ", "")
    version = pkg_resources.get_distribution("ffun").version
    return f"{name}/{version} ({settings.environment} {settings.app_domain})"



import datetime
import enum
import uuid
from typing import Any, Iterable

import markdown
import pydantic

from ffun.core import api
from ffun.core.entities import BaseEntity
from ffun.feeds import entities as f_entities
from ffun.feeds_collections import entities as fc_entities
from ffun.feeds_links import entities as fl_entities
from ffun.library import entities as l_entities
from ffun.markers import entities as m_entities
from ffun.ontology import entities as o_entities
from ffun.parsers import entities as p_entities
from ffun.resources import entities as r_entities
from ffun.scores import entities as s_entities
from ffun.user_settings import types as us_types
from ffun.user_settings.values import user_settings


class Marker(enum.StrEnum):
    read = "read"

    @classmethod
    def from_internal(cls, marker: m_entities.Marker) -> "Marker":
        return cls(marker.name)

    def to_internal(self) -> m_entities.Marker:
        return m_entities.Marker[self.value]


class Feed(BaseEntity):
    id: uuid.UUID
    title: str | None
    description: str | None
    url: str
    state: str
    lastError: str | None = None
    loadedAt: datetime.datetime | None
    linkedAt: datetime.datetime | None

    @classmethod
    def from_internal(cls, feed: f_entities.Feed, link: fl_entities.FeedLink) -> "Feed":
        return cls(
            id=feed.id,
            title=feed.title,
            description=feed.description,
            url=feed.url,
            state=feed.state.name,
            lastError=feed.last_error.name if feed.last_error else None,
            loadedAt=feed.loaded_at,
            linkedAt=link.created_at,
        )


class Entry(BaseEntity):
    id: uuid.UUID
    feed_id: uuid.UUID
    title: str
    url: str
    tags: list[str]
    markers: list[Marker] = []
    score: int
    scoreContributions: dict[str, int]
    publishedAt: datetime.datetime
    catalogedAt: datetime.datetime
    body: str | None = None

    @classmethod
    def from_internal(  # noqa: CFQ002
        cls,
        entry: l_entities.Entry,
        tags: Iterable[str],
        markers: Iterable[Marker],
        score: int,
        score_contributions: dict[str, int],
        with_body: bool = False,
    ) -> "Entry":
        return cls(
            id=entry.id,
            feed_id=entry.feed_id,
            title=entry.title,
            url=entry.external_url,
            tags=list(tags),
            markers=list(markers),
            score=score,
            scoreContributions=score_contributions,
            publishedAt=entry.published_at,
            catalogedAt=entry.cataloged_at,
            body=entry.body if with_body else None,
        )


class Rule(BaseEntity):
    id: uuid.UUID
    tags: list[str]
    score: int
    createdAt: datetime.datetime
    updatedAt: datetime.datetime

    @classmethod
    def from_internal(cls, rule: s_entities.Rule, tags_mapping: dict[int, str]) -> "Rule":
        return cls(
            id=rule.id,
            tags=[tags_mapping[tag_id] for tag_id in rule.tags],
            score=rule.score,
            createdAt=rule.created_at,
            updatedAt=rule.updated_at,
        )


class EntryInfo(BaseEntity):
    title: str
    body: str
    url: str | None
    published_at: datetime.datetime

    @classmethod
    def from_internal(cls, entry: p_entities.EntryInfo) -> "EntryInfo":
        return cls(title=entry.title, body=entry.body, url=entry.external_url, published_at=entry.published_at)


class FeedInfo(BaseEntity):
    url: str
    title: str
    description: str

    entries: list[EntryInfo]

    @classmethod
    def from_internal(cls, feed: p_entities.FeedInfo) -> "FeedInfo":
        return cls(
            url=feed.url,
            title=feed.title,
            description=feed.description,
            entries=[EntryInfo.from_internal(entry) for entry in feed.entries],
        )


class TagInfo(BaseEntity):
    uid: str
    name: str
    link: str | None
    categories: set[o_entities.TagCategory]

    @classmethod
    def from_internal(cls, tag: o_entities.Tag, uid: str) -> "TagInfo":
        assert tag.name is not None

        return cls(uid=uid, name=tag.name, link=tag.link, categories=tag.categories)


class UserSettingKind(enum.StrEnum):
    openai_api_key = "openai_api_key"
    openai_max_tokens_in_month = "openai_max_tokens_in_month"
    openai_hide_message_about_setting_up_key = "openai_hide_message_about_setting_up_key"
    openai_process_entries_not_older_than = "openai_process_entries_not_older_than"
    openai_allow_use_key_for_collections = "openai_allow_use_key_for_collections"

    @classmethod
    def from_internal(cls, kind: int) -> "UserSettingKind":
        from ffun.application.user_settings import UserSetting

        real_kind = UserSetting(kind)
        return UserSettingKind(real_kind.name)

    def to_internal(self) -> int:
        from ffun.application.user_settings import UserSetting

        return getattr(UserSetting, self.name)  # type: ignore


class UserSetting(BaseEntity):
    kind: UserSettingKind
    type: us_types.TypeId  # should not differ between front & back => no need to convert
    value: Any
    name: str
    description: str | None

    @classmethod
    def from_internal(cls, kind: int, value: str | int | float | bool) -> "UserSetting":
        from ffun.application.user_settings import UserSetting

        real_kind = UserSetting(kind)

        real_setting = user_settings.get(real_kind)

        assert real_setting is not None

        return cls(
            kind=UserSettingKind.from_internal(real_kind),
            type=real_setting.type.id,
            value=real_setting.type.normalize(value),
            name=real_setting.name,
            description=markdown.markdown(real_setting.description) if real_setting.description else None,
        )


class ResourceKind(enum.StrEnum):
    openai_tokens = "openai_tokens"

    @classmethod
    def from_internal(cls, kind: int) -> "ResourceKind":
        from ffun.application.resources import Resource

        real_kind = Resource(kind)
        return ResourceKind(real_kind.name)

    def to_internal(self) -> int:
        from ffun.application.resources import Resource

        return getattr(Resource, self.name)  # type: ignore


class ResourceHistoryRecord(pydantic.BaseModel):
    intervalStartedAt: datetime.datetime
    used: int
    reserved: int

    @classmethod
    def from_internal(cls, record: r_entities.Resource) -> "ResourceHistoryRecord":
        return cls(intervalStartedAt=record.interval_started_at, used=record.used, reserved=record.reserved)


##################
# Request/Response
##################


class GetFeedsRequest(api.APIRequest):
    pass


class GetFeedsResponse(api.APISuccess):
    feeds: list[Feed]


class GetLastEntriesRequest(api.APIRequest):
    period: datetime.timedelta | None = None

    @pydantic.field_validator("period")
    def validate_period(cls, v: None | datetime.timedelta) -> None | datetime.timedelta:
        if v is not None and v.total_seconds() < 0:
            raise ValueError("period must be positive")
        return v


class GetLastEntriesResponse(api.APISuccess):
    entries: list[Entry]


class GetEntriesByIdsRequest(api.APIRequest):
    ids: list[uuid.UUID]


class GetEntriesByIdsResponse(api.APISuccess):
    entries: list[Entry]


class CreateOrUpdateRuleRequest(api.APIRequest):
    tags: list[str]
    score: int


class CreateOrUpdateRuleResponse(api.APISuccess):
    pass


class DeleteRuleRequest(api.APIRequest):
    id: uuid.UUID


class DeleteRuleResponse(api.APISuccess):
    pass


class UpdateRuleRequest(api.APIRequest):
    id: uuid.UUID
    tags: list[str]
    score: int


class UpdateRuleResponse(api.APISuccess):
    pass


class GetRulesRequest(api.APIRequest):
    pass


class GetRulesResponse(api.APISuccess):
    rules: list[Rule]


class GetScoreDetailsRequest(api.APIRequest):
    entryId: uuid.UUID


class GetScoreDetailsResponse(api.APISuccess):
    rules: list[Rule]


class SetMarkerRequest(api.APIRequest):
    entryId: uuid.UUID
    marker: Marker


class SetMarkerResponse(api.APISuccess):
    pass


class RemoveMarkerRequest(api.APIRequest):
    entryId: uuid.UUID
    marker: Marker


class RemoveMarkerResponse(api.APISuccess):
    pass


class DiscoverFeedsRequest(api.APIRequest):
    url: str


class DiscoverFeedsResponse(api.APISuccess):
    feeds: list[FeedInfo]


class AddFeedRequest(api.APIRequest):
    url: str


class AddFeedResponse(api.APISuccess):
    pass


class AddOpmlRequest(api.APIRequest):
    content: str


class AddOpmlResponse(api.APISuccess):
    pass


class UnsubscribeRequest(api.APIRequest):
    feedId: uuid.UUID


class UnsubscribeResponse(api.APISuccess):
    pass


class GetFeedsCollectionsRequest(api.APIRequest):
    pass


class GetFeedsCollectionsResponse(api.APISuccess):
    collections: list[fc_entities.Collection]


class SubscribeToFeedsCollectionsRequest(api.APIRequest):
    collections: list[fc_entities.Collection]


class SubscribeToFeedsCollectionsResponse(api.APISuccess):
    pass


class GetTagsInfoRequest(api.APIRequest):
    uids: set[str]


class GetTagsInfoResponse(api.APISuccess):
    tags: dict[str, TagInfo]


class GetUserSettingsRequest(api.APIRequest):
    pass


class GetUserSettingsResponse(api.APISuccess):
    settings: list[UserSetting]


class SetUserSettingRequest(api.APIRequest):
    kind: UserSettingKind
    value: Any

    @pydantic.model_validator(mode="before")
    @classmethod
    def validate_value(cls, values: dict[str, Any]) -> dict[str, Any]:
        from ffun.application.user_settings import UserSetting

        kind = UserSettingKind(values["kind"]).to_internal()
        value = values.get("value")

        real_kind = UserSetting(kind)

        real_setting = user_settings.get(real_kind)

        values["value"] = real_setting.type.normalize(value)

        return values


class SetUserSettingResponse(api.APISuccess):
    pass


class GetResourceHistoryRequest(api.APIRequest):
    kind: ResourceKind


class GetResourceHistoryResponse(api.APISuccess):
    history: list[ResourceHistoryRecord]


class GetInfoRequest(api.APIRequest):
    pass


class GetInfoResponse(api.APISuccess):
    userId: uuid.UUID



import uuid

import pytest

from ffun.core.postgresql import transaction
from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged
from ffun.feeds.tests import make as f_make
from ffun.feeds_links.operations import add_link, get_linked_feeds, tech_merge_feeds
from ffun.users.tests import make as u_make


class TestMergeFeeds:
    @pytest.mark.asyncio
    async def test_nothing_to_merge(self, loaded_feed_id: uuid.UUID, another_loaded_feed_id: uuid.UUID) -> None:
        async with transaction() as trx:
            await tech_merge_feeds(trx, loaded_feed_id, another_loaded_feed_id)

    @pytest.mark.asyncio
    async def test_move_links(self) -> None:
        feed_1, feed_2, feed_3 = await f_make.n_feeds(3)

        user_a, user_b = await u_make.n_users(2)

        await add_link(user_a, feed_1.id)
        await add_link(user_b, feed_2.id)

        async with TableSizeNotChanged("fl_links"):
            async with transaction() as trx:
                await tech_merge_feeds(trx, from_feed_id=feed_2.id, to_feed_id=feed_3.id)

        links_a = await get_linked_feeds(user_a)
        links_b = await get_linked_feeds(user_b)

        assert {link.feed_id for link in links_a} == {feed_1.id}
        assert {link.feed_id for link in links_b} == {feed_3.id}

    @pytest.mark.asyncio
    async def test_rewrite_links(self) -> None:
        feed_1, feed_2, feed_3 = await f_make.n_feeds(3)

        user_a, user_b = await u_make.n_users(2)

        await add_link(user_a, feed_1.id)
        await add_link(user_b, feed_2.id)
        await add_link(user_b, feed_3.id)

        async with TableSizeDelta("fl_links", delta=-1):
            async with transaction() as trx:
                await tech_merge_feeds(trx, from_feed_id=feed_2.id, to_feed_id=feed_3.id)

        links_a = await get_linked_feeds(user_a)
        links_b = await get_linked_feeds(user_b)

        assert {link.feed_id for link in links_a} == {feed_1.id}
        assert {link.feed_id for link in links_b} == {feed_3.id}

    @pytest.mark.asyncio
    async def test_complex_ops(self) -> None:
        feed_1, feed_2, feed_3 = await f_make.n_feeds(3)

        user_a, user_b, user_c = await u_make.n_users(3)

        await add_link(user_a, feed_1.id)

        await add_link(user_b, feed_2.id)
        await add_link(user_b, feed_3.id)

        await add_link(user_c, feed_2.id)

        async with TableSizeDelta("fl_links", delta=-1):
            async with transaction() as trx:
                await tech_merge_feeds(trx, from_feed_id=feed_2.id, to_feed_id=feed_3.id)

        links_a = await get_linked_feeds(user_a)
        links_b = await get_linked_feeds(user_b)
        links_c = await get_linked_feeds(user_c)

        assert {link.feed_id for link in links_a} == {feed_1.id}
        assert {link.feed_id for link in links_b} == {feed_3.id}
        assert {link.feed_id for link in links_c} == {feed_3.id}



import uuid

import pytest
from structlog.testing import capture_logs

from ffun.core.postgresql import execute
from ffun.core.tests.helpers import assert_logs
from ffun.librarian import operations
from ffun.librarian.background_processors import EntriesProcessor
from ffun.librarian.tests import make
from ffun.library.tests import make as l_make
from ffun.ontology import domain as o_domain


class TestEntriesProcessors:
    @pytest.mark.asyncio
    async def test_no_entries_to_process(self, fake_entries_processor: EntriesProcessor) -> None:
        await operations.clear_processor_queue(fake_entries_processor.id)
        await make.end_processor_pointer(fake_entries_processor.id)

        with capture_logs() as logs:
            await fake_entries_processor.single_run()

        assert_logs(logs, no_entries_to_process=1)

    @pytest.mark.asyncio
    async def test_entries_more_than_concurrency(
        self, fake_entries_processor: EntriesProcessor, loaded_feed_id: uuid.UUID
    ) -> None:
        await operations.clear_processor_queue(fake_entries_processor.id)
        await make.end_processor_pointer(fake_entries_processor.id)

        entries = await l_make.n_entries(loaded_feed_id, 9)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        assert fake_entries_processor.concurrency <= len(entries)

        with capture_logs() as logs:
            await fake_entries_processor.single_run()

        assert_logs(logs, no_entries_to_process=0)

        tags = await o_domain.get_tags_for_entries(list(entries))

        for entry in entries_list[: fake_entries_processor.concurrency]:
            assert tags[entry.id] == {"fake-constant-tag-1", "fake-constant-tag-2"}

        for entry in entries_list[fake_entries_processor.concurrency :]:
            assert tags[entry.id] == set()

    @pytest.mark.asyncio
    async def test_entries_less_than_concurrency(
        self, fake_entries_processor: EntriesProcessor, loaded_feed_id: uuid.UUID
    ) -> None:
        await operations.clear_processor_queue(fake_entries_processor.id)
        await make.end_processor_pointer(fake_entries_processor.id)

        entries = await l_make.n_entries(loaded_feed_id, 2)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        assert fake_entries_processor.concurrency > len(entries)

        with capture_logs() as logs:
            await fake_entries_processor.single_run()

        assert_logs(logs, no_entries_to_process=0)

        tags = await o_domain.get_tags_for_entries(list(entries))

        for entry in entries_list:
            assert tags[entry.id] == {"fake-constant-tag-1", "fake-constant-tag-2"}

    @pytest.mark.asyncio
    async def test_unexisted_entries_in_queue(
        self, fake_entries_processor: EntriesProcessor, loaded_feed_id: uuid.UUID
    ) -> None:
        await operations.clear_processor_queue(fake_entries_processor.id)
        await make.end_processor_pointer(fake_entries_processor.id)

        entries = await l_make.n_entries(loaded_feed_id, 2)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        fake_entries_ids = [uuid.uuid4() for _ in range(3)]

        await operations.push_entries_to_processor_queue(
            execute, processor_id=fake_entries_processor.id, entry_ids=fake_entries_ids
        )

        assert fake_entries_processor.concurrency >= len(entries) + len(fake_entries_ids)

        with capture_logs() as logs:
            await fake_entries_processor.single_run()

        assert_logs(logs, no_entries_to_process=0, unexisted_entries_in_queue=1)

        tags = await o_domain.get_tags_for_entries(list(entries))

        for entry in entries_list:
            assert tags[entry.id] == {"fake-constant-tag-1", "fake-constant-tag-2"}

        entities_in_queue = await operations.get_entries_to_process(processor_id=fake_entries_processor.id, limit=100)

        assert entities_in_queue == []



import asyncio
import ssl

import anyio
import httpx
from pypika import PostgreSQLQuery
from pypika import functions as pypika_fn

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.feeds.entities import FeedError
from ffun.loader import errors
from ffun.loader.entities import ProxyState
from ffun.loader.settings import Proxy
from ffun.parsers import entities as p_entities
from ffun.parsers.domain import parse_feed

logger = logging.get_module_logger()


async def load_content(  # noqa: CFQ001, CCR001, C901 # pylint: disable=R0912, R0915
    url: str, proxy: Proxy, user_agent: str
) -> httpx.Response:
    error_code = FeedError.network_unknown

    log = logger.bind(url=url, proxy=proxy.name, function="load_content")

    try:
        log.info("loading_feed")

        headers = {"user-agent": user_agent, "accept-encoding": "br;q=1.0, gzip;q=0.9, deflate;q=0.8"}

        async with httpx.AsyncClient(proxies=proxy.url, headers=headers) as client:
            response = await client.get(url, follow_redirects=True)

    except httpx.RemoteProtocolError as e:
        message = str(e)

        if "illegal request line" in message:
            # TODO: at least part of such errors are caused by wrong HTTP protocol
            #       for example `http://gopractice.ru/feed/` tries to redirect to use HTTP/0.9
            log.warning("network_illegal_request_line")
            error_code = FeedError.network_illegal_request_line
        elif "Server disconnected without sending a response" in message:
            log.warning("network_disconection_without_response")
            error_code = FeedError.network_disconection_without_response
        elif "peer closed connection without sending complete message body (incomplete chunked read)" in message:
            log.warning("network_received_unkomplete_body")
            error_code = FeedError.network_received_unkomplete_body
        else:
            log.exception("remote_protocol_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ReadError as e:
        message = str(e)

        if message == "":
            error_code = FeedError.network_read_error
            log.warning("network_read_error")
        else:
            log.exception("unknown_read_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ConnectError as e:
        message = str(e)

        if "[Errno -2]" in message:
            log.warning("network_name_or_service_not_known")
            error_code = FeedError.network_name_or_service_not_known
        elif "[Errno -3]" in message:
            log.warning("network_temporary_failure_in_name_resolution")
            error_code = FeedError.network_temporary_failure_in_name_resolution
        elif "[Errno -5]" in message:
            log.warning("no_address_associated_with_hostname")
            error_code = FeedError.network_no_address_associated_with_hostname
        elif message == "":
            log.warning("undetected_connection_error")
            error_code = FeedError.network_undetected_connection_error
        elif message == "All connection attempts failed":
            log.warning("network_all_connection_attempts_failed")
            error_code = FeedError.network_all_connection_attempts_failed
        else:
            log.exception("connection_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except ssl.SSLError as e:
        log.warning("network_certificate_verify_failed")
        error_code = FeedError.network_ssl_connection_error
        raise errors.LoadError(feed_error_code=error_code) from e

    except ssl.SSLCertVerificationError as e:
        message = str(e)

        if "CERTIFICATE_VERIFY_FAILED" in message:
            log.warning("network_certificate_verify_failed")
            error_code = FeedError.network_certificate_verify_failed
        else:
            log.exception("ssl_cert_verification_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ConnectTimeout as e:
        log.warning("network_connect_timeout")
        error_code = FeedError.network_connection_timeout
        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ReadTimeout as e:
        log.warning("network_read_timeout")
        error_code = FeedError.network_read_timeout
        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.UnsupportedProtocol as e:
        log.warning("network_unsupported_protocol")
        error_code = FeedError.network_unsupported_protocol
        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.TooManyRedirects as e:
        log.warning("network_too_many_redirects")
        error_code = FeedError.network_too_many_redirects
        raise errors.LoadError(feed_error_code=error_code) from e

    except anyio.EndOfStream as e:
        log.warning("server_breaks_connection")
        error_code = FeedError.network_server_breaks_connection
        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ProxyError as e:
        message = str(e)

        if message.startswith("502 Could not resolve host"):
            log.warning("proxy_could_not_resolve_host")
            error_code = FeedError.proxy_could_not_resolve_host
        elif "TUN_ERR" in message and "ECONNREFUSED" in message:
            log.warning("proxy_connection_refused")
            error_code = FeedError.proxy_connection_refused
        elif "TUN_ERR" in message and "EHOSTUNREACH" in message:
            log.warning("proxy_no_route_to_host")
            error_code = FeedError.proxy_no_route_to_host
        elif "403" in message:
            log.warning("proxy_connection_403")
            error_code = FeedError.proxy_connection_403
        elif "502" in message:
            log.warning("proxy_connection_502")
            error_code = FeedError.proxy_connection_502
        else:
            log.exception("unknown_proxy_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.DecodingError as e:
        log.warning("network_decoding_error")
        error_code = FeedError.network_decoding_error
        raise errors.LoadError(feed_error_code=error_code) from e

    except Exception as e:
        log.exception("error_while_loading_feed")
        raise errors.LoadError(feed_error_code=error_code) from e

    if response.status_code != 200:
        log.warning("network_non_200_status_code", status_code=response.status_code)
        error_code = FeedError.network_non_200_status_code
        raise errors.LoadError(feed_error_code=error_code)

    log.info("feed_loaded", url=url, proxy=proxy.name)

    return response


# TODO: tests
async def decode_content(response: httpx.Response) -> str:
    error_code = FeedError.parsing_base_error

    encoding = response.encoding or "utf-8"

    try:
        return response.content.decode(encoding)
    except UnicodeDecodeError as e:
        logger.warning("unicode_decode_error_while_decoding_feed")
        error_code = FeedError.parsing_unicode_decode_error
        raise errors.LoadError(feed_error_code=error_code) from e
    except Exception as e:
        logger.exception("error_while_decoding_feed")
        raise errors.LoadError(feed_error_code=error_code) from e


# TODO: tests
async def parse_content(content: str, original_url: str) -> p_entities.FeedInfo:
    try:
        feed_info = parse_feed(content, original_url=original_url)
    except Exception as e:
        logger.exception("error_while_parsing_feed")
        raise errors.LoadError(feed_error_code=FeedError.parsing_format_error) from e

    if feed_info is None:
        raise errors.LoadError(feed_error_code=FeedError.parsing_feed_content_not_found)

    if not feed_info.entries:
        raise errors.LoadError(feed_error_code=FeedError.protocol_no_entries_in_feed)

    return feed_info


async def check_proxy(proxy: Proxy, url: str, user_agent: str) -> bool:
    try:
        async with httpx.AsyncClient(proxies=proxy.url, headers={"user-agent": user_agent}) as client:
            response = await client.head(url)
    except Exception as e:
        logger.info("proxy_check_error", proxy=proxy.name, url=url, error=str(e))
        return False

    logger.info("proxy_check", proxy=proxy.name, url=url, status_code=response.status_code)

    return response.status_code == 200


async def is_proxy_available(proxy: Proxy, anchors: list[str], user_agent: str) -> bool:
    tasks = [check_proxy(proxy, url, user_agent) for url in anchors]

    results = await asyncio.gather(*tasks, return_exceptions=True)

    return any(result for result in results)


async def get_proxy_states(names: list[str]) -> dict[str, ProxyState]:
    sql = """
    SELECT name, state
    FROM lr_proxy_states
    WHERE name = ANY(%(names)s)
    """

    rows = await execute(sql, {"names": names})

    states = {name: ProxyState.available for name in names}

    for row in rows:
        states[row["name"]] = ProxyState(row["state"])

    return states


async def update_proxy_states(states: dict[str, ProxyState]) -> None:
    if not states:
        return

    query = PostgreSQLQuery.into("lr_proxy_states").columns("name", "state", "updated_at")

    for name, state in states.items():
        query = query.insert(name, state, pypika_fn.Now())

    query = query.on_conflict("name").do_update("state").do_update("updated_at", pypika_fn.Now())

    await execute(str(query))



import json
from typing import Any

from ffun.core import json as core_json
from ffun.core import logging
from ffun.core import text as core_text
from ffun.librarian import errors
from ffun.librarian.processors import base
from ffun.library.entities import Entry
from ffun.ontology.entities import ProcessorTag
from ffun.openai import client as oai_client
from ffun.openai import errors as oai_errors
from ffun.openai.keys_rotator import api_key_for_feed_entry

logger = logging.get_module_logger()


system = """\
You are an expert on the analysis of text semantics.
For provided text, you determine a list of best tags to describe the text.
For each category, you provide 30 tags.

Categories are topics, meta-topics, high-level-topics, low-level-topics, related-topics,\
indirect-topics, mentions, indirect-mentions.

Tags are only in English. Normalize tags and output them as JSON.\
"""


# add url to allow chatGPT decide on domain
def entry_to_text(entry: Entry) -> str:
    return f'<h1>{entry.title}</h1><a href="{entry.external_url}">full article</a>{entry.body}'


trash_system_tags = {
    "topics",
    "meta-topics",
    "high-level-topics",
    "low-level-topics",
    "related-topics",
    "indirect-topics",
    "mentions",
    "indirect-mentions",
}


def extract_tags(text: str) -> set[str]:
    try:
        data = core_json.loads_with_fix(text)
        tags = core_json.extract_tags_from_random_json(data)
    except json.decoder.JSONDecodeError:
        tags = core_json.extract_tags_from_invalid_json(text)

    return tags - trash_system_tags


class Processor(base.Processor):
    __slots__ = ("api_key", "model")

    def __init__(self, model: str, **kwargs: Any):
        super().__init__(**kwargs)
        self.model = model

    async def process(self, entry: Entry) -> list[ProcessorTag]:
        tags: list[ProcessorTag] = []

        dirty_text = entry_to_text(entry)

        text = core_text.clear_text(dirty_text)

        total_tokens = 16 * 1024
        max_return_tokens = 2 * 1024

        messages = await oai_client.prepare_requests(
            system=system,
            text=text,
            model=self.model,
            function=None,
            total_tokens=total_tokens,
            max_return_tokens=max_return_tokens,
        )

        try:
            async with api_key_for_feed_entry(
                entry.feed_id, entry_age=entry.age, reserved_tokens=len(messages) * total_tokens
            ) as api_key_usage:
                results = await oai_client.multiple_requests(
                    api_key=api_key_usage.api_key,
                    model=self.model,
                    messages=messages,
                    function=None,
                    max_return_tokens=max_return_tokens,
                    temperature=0,
                    top_p=0,
                    presence_penalty=0,
                    frequency_penalty=0,
                )

                api_key_usage.used_tokens = sum(result.total_tokens for result in results)

        except oai_errors.NoKeyFoundForFeed as e:
            raise errors.SkipEntryProcessing(message=str(e)) from e

        for result in results:
            for raw_tag in extract_tags(result.content):
                tags.append(ProcessorTag(raw_uid=raw_tag))

        return tags



import uuid
from typing import Iterable

import async_lru

from ffun.feeds import domain as f_domain
from ffun.feeds import errors as f_errors
from ffun.feeds_collections.entities import Collection
from ffun.feeds_collections.predefines import predefines


def get_collections() -> Iterable[Collection]:
    return predefines.keys()


def get_feeds_for_collecton(collection: Collection) -> set[str]:
    return predefines[collection]


@async_lru.alru_cache()
async def is_feed_in_collections(feed_id: uuid.UUID) -> bool:
    try:
        feed = await f_domain.get_feed(feed_id)
    except f_errors.NoFeedFound:
        return False

    for collection in get_collections():
        if feed.url in get_feeds_for_collecton(collection):
            return True

    return False



import datetime

import pydantic_settings

from ffun.core.settings import BaseSettings


class Settings(BaseSettings):
    key_quota_timeout: datetime.timedelta = datetime.timedelta(hours=1)
    key_broken_timeout: datetime.timedelta = datetime.timedelta(hours=1)

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_OPENAI_")


settings = Settings()



import asyncio
import uuid

from ffun.application.application import with_app
from ffun.cli.application import app
from ffun.core import logging
from ffun.meta import domain as m_domain

logger = logging.get_module_logger()


async def run(base_feed_id: uuid.UUID, merged_feed_id: uuid.UUID) -> None:
    async with with_app():
        await m_domain.merge_feeds(base_feed_id, merged_feed_id)


@app.command()
def merge_feeds(base_feed_id: uuid.UUID, merged_feed_id: uuid.UUID) -> None:
    asyncio.run(run(base_feed_id, merged_feed_id))



import uuid

from ffun.domain import urls as d_urls
from ffun.feeds import domain as f_domain
from ffun.library import operations
from ffun.library.entities import Entry, EntryChange

catalog_entries = operations.catalog_entries
get_entries_by_ids = operations.get_entries_by_ids
get_entries_by_filter = operations.get_entries_by_filter
check_stored_entries_by_external_ids = operations.check_stored_entries_by_external_ids
all_entries_iterator = operations.all_entries_iterator
get_entries_after_pointer = operations.get_entries_after_pointer
tech_move_entry = operations.tech_move_entry
tech_get_feed_entries_tail = operations.tech_get_feed_entries_tail
tech_remove_entries_by_ids = operations.tech_remove_entries_by_ids


async def get_entry(entry_id: uuid.UUID) -> Entry:
    entries = await get_entries_by_ids([entry_id])
    found_entry = entries.get(entry_id)
    assert found_entry is not None
    return found_entry


async def normalize_entry(entry: Entry, apply: bool = False) -> list[EntryChange]:
    feed = await f_domain.get_feed(entry.feed_id)

    new_external_url = d_urls.normalize_external_url(entry.external_url, feed.url)

    changes = []

    if new_external_url != entry.external_url:
        changes.append(
            EntryChange(id=entry.id, field="external_url", old_value=entry.external_url, new_value=new_external_url)
        )
        if apply and new_external_url is not None:
            await operations.update_external_url(entry.id, new_external_url)

    return changes



import pytest

from ffun.domain import urls


class TestNormalizeClassicUrl:
    @pytest.mark.parametrize(
        "original_url, raw_url, normalized_url",
        [
            ("https://example.com/feed/", "https://example.com", "https://example.com"),
            ("https://example.com/feed/", "https://example.com/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            (
                "https://example.com/feed/",
                "http://another.domain:666/path/a/b?c=d",
                "http://another.domain:666/path/a/b?c=d",
            ),
            (
                "https://example.com/feed/",
                "another.domain:666/path/a/b?c=d",
                "https://another.domain:666/path/a/b?c=d",
            ),
            ("https://example.com/feed/", "another.domain/path/a/b?c=d", "https://another.domain/path/a/b?c=d"),
            ("https://example.com/feed/", "/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            ("https://example.com/feed/", "path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            ("https://example.com/feed/", "path", "https://example.com/path"),
            ("https://example.com/feed/", "?c=d", "https://example.com?c=d"),
            ("https://example.com/feed/", "another.domain", "https://another.domain"),
            ("example.com/feed/", "https://example.com", "https://example.com"),
            ("example.com/feed/", "example.com/path/a/b?c=d", "example.com/path/a/b?c=d"),
            ("example.com/feed/", "http://another.domain:666/path/a/b?c=d", "http://another.domain:666/path/a/b?c=d"),
            ("example.com/feed/", "another.domain:666/path/a/b?c=d", "another.domain:666/path/a/b?c=d"),
            ("example.com/feed/", "another.domain/path/a/b?c=d", "another.domain/path/a/b?c=d"),
            ("example.com/feed/", "/path/a/b?c=d", "example.com/path/a/b?c=d"),
            ("example.com/feed/", "path/a/b?c=d", "example.com/path/a/b?c=d"),
            ("example.com/feed/", "path", "example.com/path"),
            ("example.com/feed/", "?c=d", "example.com?c=d"),
            ("example.com/feed/", "another.domain", "another.domain"),
        ],
    )
    def test_base_transformations(self, original_url: str, raw_url: str, normalized_url: str) -> None:
        assert urls.normalize_classic_url(raw_url, original_url) == normalized_url


class TestIsMagneticUrl:
    @pytest.mark.parametrize(
        "url, is_magnetic",
        [
            (
                "magnet:?xt=urn:btih:123456789abcdef0123456789abcdef0123456789&dn=Example+File.mp4&tr=udp%3A%2F%2Ftracker.example.com%3A80",  # noqa
                True,
            ),
            ("https://example.com/path/a/b?c=d", False),
            ("http://another.domain:666/path/a/b?c=d", False),
        ],
    )
    def test(self, url: str, is_magnetic: bool) -> None:
        assert urls.is_magnetic_url(url) == is_magnetic


class TestNormalizeMagneticUrl:
    def test(self) -> None:
        url = "magnet:?xt=urn:btih:123456789abcdef0123456789abcdef0123456789&dn=Example+File.mp4&tr=udp%3A%2F%2Ftracker.example.com%3A80"  # noqa
        assert urls.normalize_magnetic_url(url) == url


class TestNormalizeExternalUrl:
    @pytest.mark.parametrize(
        "url, normalized_url",
        [
            ("https://example.com/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            ("http://another.domain:666/path/a/b?c=d", "http://another.domain:666/path/a/b?c=d"),
            ("/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            (
                "magnet:?xt=urn:btih:123456789abcdef0123456789abcdef0123456789&dn=Example+File.mp4&tr=udp%3A%2F%2Ftracker.example.com%3A80",  # noqa
                "magnet:?xt=urn:btih:123456789abcdef0123456789abcdef0123456789&dn=Example+File.mp4&tr=udp%3A%2F%2Ftracker.example.com%3A80",  # noqa
            ),
            ("http://www.usinenouvelle.comhttps://www.usine-digitale.fr/article/christophe", None),
        ],
    )
    def test(self, url: str, normalized_url: str) -> None:
        assert urls.normalize_external_url(url, original_url="https://example.com") == normalized_url


class TestUrlToUid:
    @pytest.mark.parametrize(
        "url, uid",
        [  # strip spaces
            ("   example.com", "example.com"),
            ("example.com   ", "example.com"),
            # trailing slash
            ("example.com/", "example.com"),
            ("example.com//", "example.com"),
            ("example.com/path/", "example.com/path"),
            # uppercase
            ("EXAMPLE.COM", "example.com"),
            ("exaMple.com/somE/pAth", "example.com/some/path"),
            # pluses
            ("example.com/with+plus", "example.com/with+plus"),
            ("example.com/with%20space", "example.com/with+space"),
            ("example.com/with plus", "example.com/with+plus"),
            # quotes
            ("example.com/with%22quote", 'example.com/with"quote'),
            ("example.com/with%27quote", "example.com/with'quote"),
            # query
            ("example.com/?a=b&c=d", "example.com?a=b&c=d"),
            ("example.com/?c=d&a=b", "example.com?a=b&c=d"),
            # query with quotes
            ("example.com/?a=%22b%22", 'example.com?a="b"'),
            ("example.com/?a=%27b%27", "example.com?a='b'"),
            # complext query with duplicates and sorting
            ("example.com/?c=d&a=g&e=f&a=b", "example.com?a=b&a=g&c=d&e=f"),
            # duble slashes
            ("example.com//path//to//resource", "example.com/path/to/resource"),
            # unicode normalization
            ("example.com/%D0%BF%D1%80%D0%B8%D0%B2%D0%B5%D1%82", "example.com/привет"),
            # unicode complex symbols
            ("example.com/%E2%98%83", "example.com/☃"),
            # unicode NFD to NFC
            ("example.com/й", "example.com/й"),
            # remove schema
            ("http://example.com", "example.com"),
            ("https://example.com", "example.com"),
            ("ftp://example.com", "example.com"),
            ("//example.com", "example.com"),
            # remove ports
            ("example.com:666", "example.com"),
            ("example.com:666/path", "example.com/path"),
            ("example.com:80", "example.com"),
            ("example.com:443/path", "example.com/path"),
            # remove fragment
            ("example.com#fragment", "example.com"),
            ("example.com/path#fragment", "example.com/path"),
            ("example.com/path#fragment?query", "example.com/path"),
            ("example.com/path?query#", "example.com/path?query"),
        ],
    )
    def test(self, url: str, uid: str) -> None:
        assert urls.url_to_uid(url) == uid



import datetime
import uuid

from ffun.core.entities import BaseEntity


class ProcessorPointer(BaseEntity):
    processor_id: int
    pointer_created_at: datetime.datetime
    pointer_entry_id: uuid.UUID



import yarl
from bs4 import BeautifulSoup

from ffun.core import logging
from ffun.loader import domain as lo_domain
from ffun.loader import errors as lo_errors
from ffun.parsers import entities as p_entities

logger = logging.get_module_logger()


def fix_url(raw_url: str) -> str:
    url = yarl.URL(raw_url)

    if url.is_absolute():
        if not url.scheme:
            url = url.with_scheme("http")

    else:
        url = yarl.URL("http://" + raw_url)

    return url.human_repr()


async def extract_content(url: str) -> str | None:
    logger.info("extract_content")

    try:
        response = await lo_domain.load_content_with_proxies(url)
        content = await lo_domain.decode_content(response)
    except lo_errors.LoadError:
        logger.info("can_not_access_content")
        return None
    except Exception:
        logger.exception("unexpected_error_while_parsing_feed")
        return None

    return content


async def extract_feed_info(content: str, original_url: str) -> p_entities.FeedInfo | None:
    logger.info("extract_feed_info")

    try:
        return await lo_domain.parse_content(content, original_url=original_url)
    except lo_errors.LoadError:
        logger.info("feed_not_found_at_start_url")
    except Exception:
        logger.exception("unexpected_error_while_parsing_feed")

    return None


def construct_soup(content: str) -> BeautifulSoup | None:
    logger.info("construct_soup")

    try:
        return BeautifulSoup(content, "html.parser")
    except Exception:
        logger.exception("unexpected_error_while_parsing_html")

    return None


async def extract_feeds_from_links(soup: BeautifulSoup) -> list[p_entities.FeedInfo]:
    results = []
    links_to_check = set()

    for link in soup("link"):
        if link.has_attr("href"):
            links_to_check.add(link["href"])

    logger.info("links_to_check", links_to_check=links_to_check)

    for link in links_to_check:
        results.extend(await discover(url=link, stop=True))

    logger.info("results", result_links=[feed_info.url for feed_info in results])

    return results


async def extract_feeds_from_a(soup: BeautifulSoup) -> list[p_entities.FeedInfo]:
    results = []
    links_to_check = set()

    # TODO: can be very-very long, must be improved
    for link in list(soup("a")):
        if link.has_attr("href"):
            links_to_check.add(link["href"])

    logger.info("links_to_check", links_to_check=links_to_check)

    for link in links_to_check:
        results.extend(await discover(url=link, stop=True))

    logger.info("results", result_links=[feed_info.url for feed_info in results])

    return results


# TODO: this is very straightforward implementation, it should be improved
# TODO: check already saved feeds, but remember, that there may be saved only part of the feeds from url
@logging.bound_function()
async def discover(url: str, stop: bool = False) -> list[p_entities.FeedInfo]:
    url = fix_url(url)

    logger.info("fixed_url", url=url)

    logger.info("start_discover")

    content = await extract_content(url)

    if content is None:
        return []

    feed_info = await extract_feed_info(content, original_url=url)

    if feed_info is not None:
        return [feed_info]

    if stop:
        return []

    logger.info("extract_feeds_candidates")

    soup = construct_soup(content)

    if soup is None:
        return []

    logger.info("search_link_candidates")

    results = await extract_feeds_from_links(soup)

    if results:
        # Here we expect "correct" sites behavior "if there are links to feeds, then they will contain all feeds"
        # it is not always be true, but it is good enough for now
        return results

    results = await extract_feeds_from_links(soup)

    return results


async def check_if_feed(url: str) -> p_entities.FeedInfo | None:
    feeds = await discover(url=url, stop=True)

    if not feeds:
        return None

    return feeds[0]



import datetime
import uuid
from typing import Any, AsyncGenerator, Iterable

import psycopg

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.library import errors
from ffun.library.entities import Entry

logger = logging.get_module_logger()


sql_insert_entry = """
INSERT INTO l_entries (id, feed_id, title, body,
                       external_id, external_url, external_tags, published_at)
VALUES (%(id)s, %(feed_id)s, %(title)s, %(body)s,
        %(external_id)s, %(external_url)s, %(external_tags)s, %(published_at)s)
"""


def row_to_entry(row: dict[str, Any]) -> Entry:
    row["cataloged_at"] = row.pop("created_at")
    return Entry(**row)


async def catalog_entries(entries: Iterable[Entry]) -> None:
    for entry in entries:
        try:
            await execute(
                sql_insert_entry,
                {
                    "id": entry.id,
                    "feed_id": entry.feed_id,
                    "title": entry.title,
                    "body": entry.body,
                    "external_id": entry.external_id,
                    "external_url": entry.external_url,
                    "external_tags": list(entry.external_tags),
                    "published_at": entry.published_at,
                },
            )
        except psycopg.errors.UniqueViolation:
            logger.warning("racing_is_possible_unique_violation_while_saving_entry", entry_id=entry.id)


# we must controll unique constraint on entries by pair feed + id because
# 1. we can not guarantee that there will be no duplicates of feeds
#    (can not guarantee perfect normalization of urls, etc)
# 2. someone can damage database by infecting with wrong entries from faked feed
async def check_stored_entries_by_external_ids(feed_id: uuid.UUID, external_ids: Iterable[str]) -> set[str]:
    sql = """
    SELECT external_id
    FROM l_entries
    WHERE feed_id = %(feed_id)s AND external_id = ANY(%(external_ids)s)
    """

    rows = await execute(sql, {"external_ids": external_ids, "feed_id": feed_id})

    return {row["external_id"] for row in rows}


sql_select_entries = """SELECT * FROM l_entries WHERE id = ANY(%(ids)s)"""


async def get_entries_by_ids(ids: Iterable[uuid.UUID]) -> dict[uuid.UUID, Entry | None]:
    rows = await execute(sql_select_entries, {"ids": ids})

    result: dict[uuid.UUID, Entry | None] = {id: None for id in ids}

    for row in rows:
        result[row["id"]] = row_to_entry(row)

    return result


async def get_entries_by_filter(
    feeds_ids: Iterable[uuid.UUID], limit: int, period: datetime.timedelta | None = None
) -> list[Entry]:
    if period is None:
        period = datetime.timedelta(days=100 * 365)

    sql = """
    SELECT * FROM l_entries
    WHERE created_at > NOW() - %(period)s and feed_id = ANY(%(feeds_ids)s)
    ORDER BY created_at DESC
    LIMIT %(limit)s"""

    rows = await execute(sql, {"feeds_ids": feeds_ids, "period": period, "limit": limit})

    return [row_to_entry(row) for row in rows]


async def get_entries_after_pointer(
    created_at: datetime.datetime, entry_id: uuid.UUID, limit: int
) -> list[tuple[uuid.UUID, datetime.datetime]]:
    # Indenx on created_at should be enough (currently it is (created_at, feed_idid))
    # In case of troubles we could add index on (created_at, id)
    sql = """
    SELECT id, created_at FROM l_entries
    WHERE created_at > %(created_at)s OR
          (created_at = %(created_at)s AND id > %(entry_id)s)
    ORDER BY created_at ASC, id ASC
    LIMIT %(limit)s
    """

    rows = await execute(sql, {"created_at": created_at, "entry_id": entry_id, "limit": limit})

    return [(row["id"], row["created_at"]) for row in rows]


# iterate by pairs (feed_id, entry_id) because we already have index on it
# TODO: rewrite to use get_entries_after_pointer
async def all_entries_iterator(chunk: int) -> AsyncGenerator[Entry, None]:
    feed_id = uuid.UUID("00000000-0000-0000-0000-000000000000")
    entry_id = uuid.UUID("00000000-0000-0000-0000-000000000000")

    sql = """
    SELECT * FROM l_entries
    WHERE (%(feed_id)s, %(entry_id)s) < (feed_id, id)
    ORDER BY feed_id, id ASC
    LIMIT %(chunk)s
    """

    while True:
        rows = await execute(sql, {"feed_id": feed_id, "entry_id": entry_id, "chunk": chunk})

        if not rows:
            break

        for row in rows:
            yield row_to_entry(row)

        feed_id = rows[-1]["feed_id"]
        entry_id = rows[-1]["id"]


async def update_external_url(entity_id: uuid.UUID, url: str) -> None:
    sql = """
    UPDATE l_entries
    SET external_url = %(url)s
    WHERE id = %(entity_id)s
    """

    await execute(sql, {"entity_id": entity_id, "url": url})


async def tech_remove_entries_by_ids(entries_ids: Iterable[uuid.UUID]) -> None:
    sql = """
    DELETE FROM l_entries
    WHERE id = ANY(%(entries_ids)s)
    """

    await execute(sql, {"entries_ids": list(entries_ids)})


async def tech_move_entry(entry_id: uuid.UUID, feed_id: uuid.UUID) -> None:
    sql = """
    UPDATE l_entries
    SET feed_id = %(feed_id)s
    WHERE id = %(entry_id)s
    """

    try:
        await execute(sql, {"entry_id": entry_id, "feed_id": feed_id})
    except psycopg.errors.UniqueViolation as e:
        raise errors.CanNotMoveEntryAlreadyInFeed(entry_id=entry_id, feed_id=feed_id) from e


async def tech_get_feed_entries_tail(feed_id: uuid.UUID, offset: int) -> set[uuid.UUID]:
    """
    Get the last entries for the feed.
    """
    # Order by published_at because we want to keep the newest entries
    # and it is better to take decission based on time from an entry's source rather than on time when we collected it
    sql = """
    SELECT id
    FROM l_entries
    WHERE feed_id = %(feed_id)s
    ORDER BY published_at DESC
    OFFSET %(offset)s
    """

    result = await execute(sql, {"feed_id": feed_id, "offset": offset})

    return {row["id"] for row in result}



import uuid

import pytest
from pytest_mock import MockerFixture
from structlog.testing import capture_logs

from ffun.core.tests.helpers import assert_logs
from ffun.feeds import domain as f_domain
from ffun.feeds import entities as f_entities
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.library import entities as l_entities
from ffun.loader.domain import check_proxies_availability, detect_orphaned, process_feed, store_entries, sync_feed_info
from ffun.loader.settings import Proxy, settings
from ffun.parsers import entities as p_entities
from ffun.parsers.tests import make as p_make


def assert_entriy_equal_to_info(entry_info: p_entities.EntryInfo, entry: l_entities.Entry) -> None:
    assert entry.title == entry_info.title
    assert entry.body == entry_info.body
    assert entry.external_id == entry_info.external_id
    assert entry.external_url == entry_info.external_url
    assert entry.external_tags == entry_info.external_tags
    assert entry.published_at == entry_info.published_at


class TestDetectOrphaned:
    @pytest.mark.asyncio
    async def test_is_orphaned(self, saved_feed_id: uuid.UUID) -> None:
        assert await detect_orphaned(saved_feed_id)

        loaded_feed = await f_domain.get_feed(saved_feed_id)

        assert loaded_feed.state == f_entities.FeedState.orphaned

    @pytest.mark.asyncio
    async def test_is_orphaned_but_from_collection(self, saved_collection_feed_id: uuid.UUID) -> None:
        assert not await detect_orphaned(saved_collection_feed_id)

        loaded_feed = await f_domain.get_feed(saved_collection_feed_id)

        assert loaded_feed.state != f_entities.FeedState.orphaned

    @pytest.mark.asyncio
    async def test_not_orphaned(self, internal_user_id: uuid.UUID, saved_feed_id: uuid.UUID) -> None:
        await fl_domain.add_link(internal_user_id, saved_feed_id)

        assert not await detect_orphaned(saved_feed_id)

        loaded_feed = await f_domain.get_feed(saved_feed_id)

        assert loaded_feed.state != f_entities.FeedState.orphaned


class TestSyncFeedInfo:
    @pytest.mark.asyncio
    async def test_no_sync_required(self, saved_feed: f_entities.Feed, mocker: MockerFixture) -> None:
        update_feed_info = mocker.patch("ffun.feeds.domain.update_feed_info")

        assert saved_feed.title
        assert saved_feed.description

        feed_info = p_entities.FeedInfo(
            url=saved_feed.url, title=saved_feed.title, description=saved_feed.description, entries=[]
        )

        await sync_feed_info(saved_feed, feed_info)

        update_feed_info.assert_not_called()

    @pytest.mark.asyncio
    async def test_sync_required(self, saved_feed: f_entities.Feed) -> None:
        feed_info = p_entities.FeedInfo(
            url=saved_feed.url, title=uuid.uuid4().hex, description=uuid.uuid4().hex, entries=[]
        )

        await sync_feed_info(saved_feed, feed_info)

        loaded_feed = await f_domain.get_feed(saved_feed.id)

        assert loaded_feed.title == feed_info.title
        assert loaded_feed.description == feed_info.description


class TestStoreEntries:
    @pytest.mark.asyncio
    async def test_no_entries(self, saved_feed_id: uuid.UUID) -> None:
        await store_entries(saved_feed_id, [])

        entries = await l_domain.get_entries_by_filter([saved_feed_id], limit=1)

        assert not entries

    @pytest.mark.asyncio
    async def test_save_new_entries(self, saved_feed_id: uuid.UUID) -> None:
        n = 3

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]

        await store_entries(saved_feed_id, entry_infos)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed_id], limit=n + 1)

        assert len(loaded_entries) == 3

        entry_infos.sort(key=lambda e: e.title)
        loaded_entries.sort(key=lambda e: e.title)

        for entry_info, entry in zip(entry_infos, loaded_entries):
            assert entry.feed_id == saved_feed_id
            assert_entriy_equal_to_info(entry_info, entry)

    @pytest.mark.asyncio
    async def test_save_in_parts(self, saved_feed_id: uuid.UUID) -> None:
        n = 5
        m = 3

        assert m < n

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]
        entry_infos.sort(key=lambda e: e.title)

        await store_entries(saved_feed_id, entry_infos[:m])

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed_id], limit=n + 1)

        assert len(loaded_entries) == m

        loaded_entries.sort(key=lambda e: e.title)

        for entry_info, entry in zip(entry_infos, loaded_entries):
            assert entry.feed_id == saved_feed_id
            assert_entriy_equal_to_info(entry_info, entry)

        await store_entries(saved_feed_id, entry_infos)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed_id], limit=n + 1)

        assert len(loaded_entries) == n

        loaded_entries.sort(key=lambda e: e.title)

        for entry_info, entry in zip(entry_infos, loaded_entries):
            assert entry.feed_id == saved_feed_id
            assert_entriy_equal_to_info(entry_info, entry)


class TestProcessFeed:
    @pytest.mark.asyncio
    async def test_orphaned_feed(self, saved_feed: f_entities.Feed) -> None:
        assert await detect_orphaned(saved_feed.id)

        await process_feed(feed=saved_feed)

        loaded_feed = await f_domain.get_feed(saved_feed.id)

        assert loaded_feed.state == f_entities.FeedState.orphaned

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=1)

        assert not loaded_entries

    @pytest.mark.asyncio
    async def test_can_not_extract_feed(
        self, internal_user_id: uuid.UUID, saved_feed: f_entities.Feed, mocker: MockerFixture
    ) -> None:
        extract_feed_info = mocker.patch("ffun.loader.domain.extract_feed_info", return_value=None)

        await fl_domain.add_link(internal_user_id, saved_feed.id)

        await process_feed(feed=saved_feed)

        extract_feed_info.assert_called_once_with(saved_feed)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=1)

        assert not loaded_entries

    @pytest.mark.asyncio
    async def test_success(
        self, internal_user_id: uuid.UUID, saved_feed: f_entities.Feed, mocker: MockerFixture
    ) -> None:
        n = 3

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]
        entry_infos.sort(key=lambda e: e.title)

        assert saved_feed.title
        assert saved_feed.description

        feed_info = p_entities.FeedInfo(
            url=saved_feed.url, title=saved_feed.title, description=saved_feed.description, entries=entry_infos
        )

        extract_feed_info = mocker.patch("ffun.loader.domain.extract_feed_info", return_value=feed_info)

        await fl_domain.add_link(internal_user_id, saved_feed.id)

        with capture_logs() as logs:
            await process_feed(feed=saved_feed)

        assert_logs(logs, feed_has_no_entries_tail=1, feed_entries_tail_removed=0)

        extract_feed_info.assert_called_once_with(saved_feed)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=n + 1)

        assert len(loaded_entries) == n

        entry_infos.sort(key=lambda e: e.title)
        loaded_entries.sort(key=lambda e: e.title)

        for entry_info, entry in zip(entry_infos, loaded_entries):
            assert entry.feed_id == saved_feed.id
            assert_entriy_equal_to_info(entry_info, entry)

    @pytest.mark.asyncio
    async def test_remove_too_long_entries_tail(
        self, internal_user_id: uuid.UUID, saved_feed: f_entities.Feed, mocker: MockerFixture
    ) -> None:
        n = 5
        m = 3

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]
        entry_infos.sort(key=lambda e: e.title)

        assert saved_feed.title
        assert saved_feed.description

        feed_info = p_entities.FeedInfo(
            url=saved_feed.url, title=saved_feed.title, description=saved_feed.description, entries=entry_infos
        )

        mocker.patch("ffun.loader.domain.extract_feed_info", return_value=feed_info)
        mocker.patch("ffun.meta.settings.settings.max_entries_per_feed", m)

        await fl_domain.add_link(internal_user_id, saved_feed.id)

        with capture_logs() as logs:
            await process_feed(feed=saved_feed)

        assert_logs(logs, feed_has_no_entries_tail=0, feed_entries_tail_removed=1)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=n + 1)

        assert len(loaded_entries) == m


class TestCheckProxiesAvailability:
    @pytest.mark.asyncio
    async def test_no_proxies(self, mocker: MockerFixture) -> None:
        mocker.patch("ffun.loader.settings.settings.proxies", [])

        is_proxy_available = mocker.patch("ffun.loader.operations.is_proxy_available")

        await check_proxies_availability()

        is_proxy_available.assert_not_called()

    @pytest.mark.asyncio
    async def test_proxies(self, mocker: MockerFixture) -> None:
        n = 3

        proxies = [Proxy(name=uuid.uuid4().hex, url=None) for _ in range(n)]

        mocker.patch("ffun.loader.settings.settings.proxies", proxies)

        is_proxy_available = mocker.patch("ffun.loader.operations.is_proxy_available")

        await check_proxies_availability()

        for proxy in proxies:
            is_proxy_available.assert_any_call(proxy=proxy, anchors=settings.proxy_anchors, user_agent="unknown")



import uuid
from typing import Iterable

from bidict import bidict

from ffun.core.postgresql import ExecuteType, execute, run_in_transaction, transaction
from ffun.ontology import operations
from ffun.ontology.entities import ProcessorTag, Tag, TagCategory, TagPropertyType
from ffun.tags import converters

_tags_cache: bidict[str, int] = bidict()


async def get_id_by_uid(tag: str) -> int:
    if tag in _tags_cache:
        return _tags_cache[tag]

    tag_id = await operations.get_or_create_id_by_tag(tag)

    _tags_cache[tag] = tag_id

    return tag_id


async def get_ids_by_uids(tags: Iterable[str]) -> dict[str, int]:
    return {tag: await get_id_by_uid(tag) for tag in tags}


async def get_tags_by_ids(ids: Iterable[int]) -> dict[int, str]:
    result = {}

    tags_to_request = []

    for tag_id in ids:
        if tag_id in _tags_cache.inverse:
            result[tag_id] = _tags_cache.inverse[tag_id]
        else:
            tags_to_request.append(tag_id)

    if not tags_to_request:
        return result

    missed_tags = await operations.get_tags_by_ids(tags_to_request)

    _tags_cache.inverse.update(missed_tags)

    result.update(missed_tags)

    return result


async def apply_tags_to_entry(entry_id: uuid.UUID, processor_id: int, tags: Iterable[ProcessorTag]) -> None:
    raw_to_uids = {tag.raw_uid: converters.normalize(tag.raw_uid) for tag in tags}

    uids_to_ids = await get_ids_by_uids(raw_to_uids.values())

    properties = []

    for tag in tags:
        properties.extend(tag.build_properties_for(uids_to_ids[raw_to_uids[tag.raw_uid]], processor_id=processor_id))

    async with transaction() as execute:
        await operations.apply_tags(execute, entry_id, processor_id, uids_to_ids.values())
        await operations.apply_tags_properties(execute, properties)


async def get_tags_ids_for_entries(entries_ids: list[uuid.UUID]) -> dict[uuid.UUID, set[int]]:
    return await operations.get_tags_for_entries(execute, entries_ids)


async def get_tags_for_entries(entries_ids: list[uuid.UUID]) -> dict[uuid.UUID, set[str]]:
    tags_ids = await operations.get_tags_for_entries(execute, entries_ids)

    all_tags = set()

    for tags in tags_ids.values():
        all_tags.update(tags)

    tags_mapping = await get_tags_by_ids(all_tags)

    result = {entry_id: {tags_mapping[tag_id] for tag_id in tags} for entry_id, tags in tags_ids.items()}

    for entry_id in entries_ids:
        if entry_id not in result:
            result[entry_id] = set()

    return result


async def get_tags_info(tags_ids: Iterable[int]) -> dict[int, Tag]:  # noqa: CCR001
    # we expect that properties will be sorted by date from the newest to the oldest
    properties = await operations.get_tags_properties(tags_ids)

    tags_by_ids = await get_tags_by_ids(tags_ids)

    info = {tag_id: Tag(id=tag_id, name=converters.verbose(tags_by_ids[tag_id])) for tag_id in tags_ids}

    # TODO: implement more complex merging
    for property in properties:
        tag = info[property.tag_id]

        if property.type == TagPropertyType.link:
            if tag.link is None:
                tag.link = property.value
            continue

        if property.type == TagPropertyType.categories:
            tag.categories.update(TagCategory(cat) for cat in property.value.split(","))
            continue

        raise NotImplementedError(f"Unknown property type: {property.type}")

    return info


@run_in_transaction
async def remove_relations_for_entries(execute: ExecuteType, entries_ids: list[uuid.UUID]) -> None:
    await operations.remove_relations_for_entries(execute, entries_ids)


@run_in_transaction
async def tech_copy_relations(execute: ExecuteType, entry_from_id: uuid.UUID, entry_to_id: uuid.UUID) -> None:
    await operations.tech_copy_relations(execute, entry_from_id, entry_to_id)



import decimal
import enum

from ffun.user_settings import types
from ffun.user_settings.values import Value, user_settings


class UserSetting(enum.IntEnum):
    openai_api_key = 1
    openai_max_tokens_in_month = 2
    openai_hide_message_about_setting_up_key = 3
    openai_process_entries_not_older_than = 4
    openai_allow_use_key_for_collections = 5


description_openai_api_key = """
Feeds Fun uses OpenAI ChatGPT to find tags for texts.

Because, for now, our service is free to use and OpenAI API costs money, \
we politely ask you to set up your own OpenAI API key.

Here's how your key will be used:

- Your key will be used only to process your feeds. We will not use it for any other purposes.
- You can establish a limit on the maximum number of tokens that can be used in a month. \
This allows you to regulate your monthly spendings on the OpenAI API.
- If multiple users are subscribed to a single feed, we'll use a key with fewer usages in the current month.
- If a user who lacks a key is subscribed to a feed and the feed's news already have tags, \
the user will see these tags.
- You can find API key usage statistics at this page.

The more users set up the key, the cheaper it will be for everyone.
"""

openai_max_token_cost = decimal.Decimal("0.004")
openai_max_token_cost_n = 1000
openai_max_spendings = decimal.Decimal("10.00")

description_openai_max_tokens_in_month = f"""
Tokens are the currency of the OpenAI API world. The more tokens you use, the more you've gotta pay.

The default limit is calculated based on an estimation that should prevent your monthly spending \
from exceeding ${openai_max_spendings}. However, this figure is merely a projection and actual usage may vary.

"""


description_openai_process_entries_not_older_than = """
Some feeds keep all their news, regardless of their age. If you subscribe to such a feed, \
it may eat a lot of your OpenAI tokens.

To prevent this, we allow limiting the age of news to be processed with your OpenAI key.

If you want to help us and tag everything, you can set this value to a big number, like 100500.
"""

description_openai_allow_use_key_for_collections = """
Enable if you want to help new users by providing good tags for feeds in standard collections. \
If you enable this your OpenAI key will be used for feeds in collections the same way as for your feeds."
"""


user_settings.add(
    Value(
        key=UserSetting.openai_api_key,
        name="OpenAI API key",
        type=types.Secret(),
        default="",
        description=description_openai_api_key,
    )
)

user_settings.add(
    Value(
        key=UserSetting.openai_max_tokens_in_month,
        name="OpenAI max tokens in month",
        type=types.Integer(),
        default=int(openai_max_spendings / openai_max_token_cost * openai_max_token_cost_n),
        description=description_openai_max_tokens_in_month,
    )
)

user_settings.add(
    Value(
        key=UserSetting.openai_hide_message_about_setting_up_key,
        name="Hide message about setting up OpenAI API key",
        type=types.Boolean(),
        default=False,
        description=None,
    )
)

user_settings.add(
    Value(
        key=UserSetting.openai_process_entries_not_older_than,
        name="Use OpenAI key only for entries not older than N days",
        type=types.Integer(),
        default=3,
        description=description_openai_process_entries_not_older_than,
    )
)

user_settings.add(
    Value(
        key=UserSetting.openai_allow_use_key_for_collections,
        name="Allow using OpenAI key for feeds in standard collections",
        type=types.Boolean(),
        default=False,
        description=description_openai_allow_use_key_for_collections,
    )
)



import uuid

import psycopg

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.users import errors
from ffun.users.entities import Service

logger = logging.get_module_logger()


async def add_mapping(service: Service, external_id: str) -> uuid.UUID:
    sql = """
        INSERT INTO u_mapping (service_id, external_id, internal_id)
        VALUES (%(service_id)s, %(external_id)s, %(internal_id)s)
    """

    internal_id = uuid.uuid4()

    try:
        await execute(sql, {"service_id": service, "external_id": external_id, "internal_id": internal_id})
    except psycopg.errors.UniqueViolation:
        return await get_mapping(service, external_id)

    return internal_id


async def get_mapping(service: Service, external_id: str) -> uuid.UUID:
    sql = """
        SELECT internal_id
        FROM u_mapping
        WHERE service_id = %(service_id)s
              AND external_id = %(external_id)s
    """

    result = await execute(sql, {"service_id": service, "external_id": external_id})

    if not result:
        raise errors.NoUserMappingFound(service=service, external_id=external_id)

    assert isinstance(result[0]["internal_id"], uuid.UUID)

    return result[0]["internal_id"]



import uuid
from typing import Iterable

from bidict import bidict

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute
from ffun.ontology.entities import TagProperty, TagPropertyType

logger = logging.get_module_logger()


async def get_tags_mappig() -> bidict[str, int]:
    sql = "SELECT id, uid FROM o_tags"
    rows = await execute(sql)
    return bidict.bidict({row["uid"]: row["id"] for row in rows})  # type: ignore


async def get_id_by_tag(tag: str) -> int | None:
    sql = "SELECT id FROM o_tags WHERE uid = %(tag)s"
    rows = await execute(sql, {"tag": tag})

    if not rows:
        return None

    assert isinstance(rows[0]["id"], int)

    return rows[0]["id"]


async def register_tag(tag: str) -> int:
    sql = """
    INSERT INTO o_tags (uid)
    VALUES (%(tag)s)
    ON CONFLICT (uid) DO NOTHING
    RETURNING id"""

    rows = await execute(sql, {"tag": tag})

    if not rows:
        tag_id = await get_id_by_tag(tag)

        assert tag_id is not None
        assert isinstance(tag_id, int)

        return tag_id

    assert isinstance(rows[0]["id"], int)

    return rows[0]["id"]


async def get_or_create_id_by_tag(tag: str) -> int:
    tag_id = await get_id_by_tag(tag)

    if tag_id is not None:
        return tag_id

    return await register_tag(tag)


async def get_tags_by_ids(tags_ids: list[int]) -> dict[int, str]:
    sql = "SELECT * FROM o_tags WHERE id = ANY(%(tags_ids)s)"
    rows = await execute(sql, {"tags_ids": tags_ids})
    return {row["id"]: row["uid"] for row in rows}


async def _save_tags(execute: ExecuteType, entry_id: uuid.UUID, tags_ids: Iterable[int]) -> None:
    sql_relations = """
    INSERT INTO o_relations (entry_id, tag_id)
    VALUES (%(entry_id)s, %(tag_id)s)
    ON CONFLICT (entry_id, tag_id) DO NOTHING"""

    for tag_id in tags_ids:
        await execute(sql_relations, {"entry_id": entry_id, "tag_id": tag_id})


async def _register_relations_processors(
    execute: ExecuteType, relations_ids: Iterable[int], processor_id: int
) -> None:
    sql_register_processor = """
    INSERT INTO o_relations_processors (relation_id, processor_id)
    VALUES (%(relation_id)s, %(processor_id)s)
    ON CONFLICT (relation_id, processor_id) DO NOTHING"""

    for relation_id in relations_ids:
        await execute(sql_register_processor, {"relation_id": relation_id, "processor_id": processor_id})


async def _get_relations_for_entry_and_tags(
    execute: ExecuteType, entry_id: uuid.UUID, tags_ids: Iterable[int]
) -> dict[int, int]:
    result = await execute(
        "SELECT id, tag_id FROM o_relations WHERE entry_id = %(entry_id)s AND tag_id = ANY(%(tags_ids)s)",
        {"entry_id": entry_id, "tags_ids": list(tags_ids)},
    )

    return {row["tag_id"]: row["id"] for row in result}


async def apply_tags(execute: ExecuteType, entry_id: uuid.UUID, processor_id: int, tags_ids: Iterable[int]) -> None:
    await _save_tags(execute, entry_id, tags_ids)

    relations = await _get_relations_for_entry_and_tags(execute, entry_id, tags_ids)

    await _register_relations_processors(execute, list(relations.values()), processor_id)


async def tech_copy_relations(execute: ExecuteType, entry_from_id: uuid.UUID, entry_to_id: uuid.UUID) -> None:
    """Copy relations with processors info."""
    # get processors for each tag in entry_from
    sql = """
    SELECT rp.processor_id, r.tag_id
    FROM o_relations_processors rp
    JOIN o_relations r ON rp.relation_id = r.id
    WHERE r.entry_id = %(entry_id)s
    """

    result = await execute(sql, {"entry_id": entry_from_id})

    tags_by_processors: dict[int, list[int]] = {}

    for row in result:
        if row["processor_id"] not in tags_by_processors:
            tags_by_processors[row["processor_id"]] = []

        tags_by_processors[row["processor_id"]].append(row["tag_id"])

    for processor_id, tags_ids in tags_by_processors.items():
        await apply_tags(execute, entry_to_id, processor_id, tags_ids)


async def apply_tags_properties(execute: ExecuteType, properties: Iterable[TagProperty]) -> None:
    sql = """
    INSERT INTO o_tags_properties (tag_id, type, value, processor_id, created_at)
    VALUES (%(tag_id)s, %(type)s, %(value)s, %(processor_id)s, %(created_at)s)
    ON CONFLICT (tag_id, type, processor_id) DO UPDATE SET value = %(value)s
    """

    for property in properties:
        await execute(
            sql,
            {
                "tag_id": property.tag_id,
                "type": property.type,
                "value": property.value,
                "processor_id": property.processor_id,
                "created_at": property.created_at,
            },
        )


async def get_tags_for_entries(execute: ExecuteType, entries_ids: list[uuid.UUID]) -> dict[uuid.UUID, set[int]]:
    sql = """SELECT * FROM o_relations WHERE entry_id = ANY(%(entries_ids)s)"""

    rows = await execute(sql, {"entries_ids": entries_ids})

    result: dict[uuid.UUID, set[int]] = {}

    for row in rows:
        entry_id = row["entry_id"]
        tag_id = row["tag_id"]

        if entry_id not in result:
            result[entry_id] = set()

        result[entry_id].add(tag_id)

    return result


async def get_tags_properties(tags_ids: Iterable[int]) -> list[TagProperty]:
    result = await execute(
        "SELECT * FROM o_tags_properties WHERE tag_id = ANY(%(tags_ids)s) ORDER BY created_at DESC",
        {"tags_ids": list(tags_ids)},
    )

    return [
        TagProperty(
            tag_id=row["tag_id"],
            type=TagPropertyType(row["type"]),
            value=row["value"],
            processor_id=row["processor_id"],
            created_at=row["created_at"],
        )
        for row in result
    ]


async def remove_relations_for_entries(execute: ExecuteType, entries_ids: list[uuid.UUID]) -> None:
    sql = "SELECT id FROM o_relations WHERE entry_id = ANY(%(entries_ids)s)"

    result = await execute(sql, {"entries_ids": entries_ids})

    relations_ids = [row["id"] for row in result]

    sql = "DELETE FROM o_relations_processors WHERE relation_id = ANY(%(relations_ids)s)"

    await execute(sql, {"relations_ids": relations_ids})

    sql = "DELETE FROM o_relations WHERE entry_id = ANY(%(entries_ids)s)"

    await execute(sql, {"entries_ids": entries_ids})



from ffun.parsers import feed
from ffun.parsers.entities import FeedInfo
from ffun.parsers.feedly import extract_feeds


def parse_opml(data: str) -> list[FeedInfo]:
    return extract_feeds(data)


parse_feed = feed.parse_feed



import uuid
from typing import Iterable

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.user_settings.entities import UserSettings

logger = logging.get_module_logger()


async def save_setting(user_id: uuid.UUID, kind: int, value: str) -> None:
    sql = """
        INSERT INTO us_settings (user_id, kind, value)
        VALUES (%(user_id)s, %(kind)s, %(value)s)
        ON CONFLICT (user_id, kind)
        DO UPDATE SET value = %(value)s,
                      updated_at = NOW()
    """

    await execute(sql, {"user_id": user_id, "kind": kind, "value": value})


async def load_settings_for_users(
    user_ids: Iterable[uuid.UUID], kinds: Iterable[int]
) -> dict[uuid.UUID, UserSettings]:
    sql = """
        SELECT *
        FROM us_settings
        WHERE user_id = ANY(%(user_ids)s)
        AND kind = ANY(%(kinds)s)
    """

    result = await execute(sql, {"user_ids": user_ids, "kinds": kinds})

    values: dict[uuid.UUID, UserSettings] = {user_id: {} for user_id in user_ids}

    for row in result:
        user_id = row["user_id"]
        kind = row["kind"]
        value = row["value"]

        assert isinstance(user_id, uuid.UUID)

        values[user_id][kind] = value

    return values


async def get_users_with_setting(kind: int, value: str) -> set[uuid.UUID]:
    sql = """
        SELECT user_id
        FROM us_settings
        WHERE kind = %(kind)s
        AND value = %(value)s
    """

    result = await execute(sql, {"kind": kind, "value": value})

    return {row["user_id"] for row in result}


async def remove_setting_for_all_users(kind: int) -> None:
    sql = """
        DELETE FROM us_settings
        WHERE kind = %(kind)s
    """

    await execute(sql, {"kind": kind})



import uuid

import pytest

from ffun.core.postgresql import execute, transaction
from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged
from ffun.library.entities import Entry
from ffun.ontology.operations import (
    _get_relations_for_entry_and_tags,
    _register_relations_processors,
    _save_tags,
    apply_tags,
    remove_relations_for_entries,
    tech_copy_relations,
)
from ffun.ontology.tests.helpers import assert_has_tags


async def assert_tags_processors(entry_id: uuid.UUID, tag_processors: dict[int, set[int]]) -> None:
    relations = await _get_relations_for_entry_and_tags(execute, entry_id, list(tag_processors.keys()))

    relations_ids = list(relations.values())

    sql = "SELECT relation_id, processor_id FROM o_relations_processors WHERE relation_id = ANY(%(relations_ids)s)"

    result = await execute(sql, {"relations_ids": relations_ids})

    expected = set()

    for tag_id, processor_ids in tag_processors.items():
        for processor_id in processor_ids:
            expected.add((relations[tag_id], processor_id))

    assert expected == {(row["relation_id"], row["processor_id"]) for row in result}


class TestSaveTags:
    @pytest.mark.asyncio
    async def test_no_tags(self, cataloged_entry: Entry) -> None:
        async with TableSizeNotChanged("o_relations"):
            await _save_tags(execute, cataloged_entry.id, [])

    @pytest.mark.asyncio
    async def test(
        self, cataloged_entry: Entry, another_cataloged_entry: Entry, three_tags_ids: tuple[int, int, int]
    ) -> None:
        async with TableSizeDelta("o_relations", delta=4):
            await _save_tags(execute, cataloged_entry.id, three_tags_ids[:2])
            await _save_tags(execute, another_cataloged_entry.id, three_tags_ids[1:])

        await assert_has_tags(
            {
                cataloged_entry.id: {three_tags_ids[0], three_tags_ids[1]},
                another_cataloged_entry.id: {three_tags_ids[1], three_tags_ids[2]},
            }
        )


class TestRegisterRelationsProcessors:
    @pytest.mark.asyncio
    async def test_no_relations(self, fake_processor_id: int) -> None:
        async with TableSizeNotChanged("o_relations_processors"):
            await _register_relations_processors(execute, [], fake_processor_id)

    @pytest.mark.asyncio
    async def test(self, cataloged_entry: Entry, fake_processor_id: int, three_tags_ids: tuple[int, int, int]) -> None:
        await _save_tags(execute, cataloged_entry.id, three_tags_ids[:2])

        relations = await _get_relations_for_entry_and_tags(execute, cataloged_entry.id, three_tags_ids[:2])

        async with TableSizeDelta("o_relations_processors", delta=2):
            await _register_relations_processors(
                execute, relations_ids=list(relations.values()), processor_id=fake_processor_id
            )

        await assert_tags_processors(
            entry_id=cataloged_entry.id,
            tag_processors={three_tags_ids[0]: {fake_processor_id}, three_tags_ids[1]: {fake_processor_id}},
        )

    @pytest.mark.asyncio
    async def test_duplicated(
        self, cataloged_entry: Entry, fake_processor_id: int, three_tags_ids: tuple[int, int, int]
    ) -> None:
        await _save_tags(execute, cataloged_entry.id, three_tags_ids)

        relations = await _get_relations_for_entry_and_tags(execute, cataloged_entry.id, three_tags_ids)

        relation_ids = list(relations.values())

        async with TableSizeDelta("o_relations_processors", delta=3):
            await _register_relations_processors(
                execute, relations_ids=relation_ids[:2], processor_id=fake_processor_id
            )
            await _register_relations_processors(
                execute, relations_ids=relation_ids[1:], processor_id=fake_processor_id
            )

        await assert_tags_processors(
            entry_id=cataloged_entry.id,
            tag_processors={
                three_tags_ids[0]: {fake_processor_id},
                three_tags_ids[1]: {fake_processor_id},
                three_tags_ids[2]: {fake_processor_id},
            },
        )


class TestGetRelationsForEntryAndTags:
    """Tested in other tests & code."""


class TestTechCopyRelations:
    @pytest.mark.asyncio
    async def test_no_relations(self, cataloged_entry: Entry, another_cataloged_entry: Entry) -> None:
        async with TableSizeNotChanged("o_relations"):
            await tech_copy_relations(execute, cataloged_entry.id, another_cataloged_entry.id)

    @pytest.mark.asyncio
    async def test_full_copy(
        self,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_tags_ids: tuple[int, int, int],
    ) -> None:
        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=fake_processor_id, tags_ids=three_tags_ids[:2]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=another_fake_processor_id, tags_ids=three_tags_ids[1:]
            )

        async with (
            TableSizeDelta("o_relations_processors", delta=4),
            TableSizeDelta("o_relations", delta=3),
            TableSizeNotChanged("o_tags"),
        ):
            async with transaction() as trx:
                await tech_copy_relations(trx, cataloged_entry.id, another_cataloged_entry.id)

        await assert_has_tags({another_cataloged_entry.id: set(three_tags_ids)})

        await assert_tags_processors(
            entry_id=another_cataloged_entry.id,
            tag_processors={
                three_tags_ids[0]: {fake_processor_id},
                three_tags_ids[1]: {fake_processor_id, another_fake_processor_id},
                three_tags_ids[2]: {another_fake_processor_id},
            },
        )

    @pytest.mark.asyncio
    async def test_some_relations_exist(
        self,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_tags_ids: tuple[int, int, int],
    ) -> None:
        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=fake_processor_id, tags_ids=three_tags_ids[:2]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=another_fake_processor_id, tags_ids=three_tags_ids[1:]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=another_cataloged_entry.id, processor_id=fake_processor_id, tags_ids=[three_tags_ids[1]]
            )

        async with (
            TableSizeDelta("o_relations_processors", delta=3),
            TableSizeDelta("o_relations", delta=2),
            TableSizeNotChanged("o_tags"),
        ):
            async with transaction() as trx:
                await tech_copy_relations(trx, cataloged_entry.id, another_cataloged_entry.id)

        await assert_has_tags({another_cataloged_entry.id: set(three_tags_ids)})

        await assert_tags_processors(
            entry_id=another_cataloged_entry.id,
            tag_processors={
                three_tags_ids[0]: {fake_processor_id},
                three_tags_ids[1]: {fake_processor_id, another_fake_processor_id},
                three_tags_ids[2]: {another_fake_processor_id},
            },
        )


class TestRemoveRelationsForEntries:
    @pytest.mark.asyncio
    async def test_nothing_to_remove(self, cataloged_entry: Entry, another_cataloged_entry: Entry) -> None:
        async with TableSizeNotChanged("o_relations"):
            await remove_relations_for_entries(execute, [cataloged_entry.id, another_cataloged_entry.id])

    @pytest.mark.asyncio
    async def test_success(
        self,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
        fake_processor_id: int,
        another_fake_processor_id: int,
        three_tags_ids: tuple[int, int, int],
    ) -> None:
        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=fake_processor_id, tags_ids=three_tags_ids[:2]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=cataloged_entry.id, processor_id=another_fake_processor_id, tags_ids=three_tags_ids[1:]
            )

        async with transaction() as trx:
            await apply_tags(
                trx, entry_id=another_cataloged_entry.id, processor_id=fake_processor_id, tags_ids=three_tags_ids
            )

        async with transaction() as trx:
            await apply_tags(
                trx,
                entry_id=another_cataloged_entry.id,
                processor_id=another_fake_processor_id,
                tags_ids=three_tags_ids,
            )

        async with (
            TableSizeDelta("o_relations_processors", delta=-4),
            TableSizeDelta("o_relations", delta=-3),
            TableSizeNotChanged("o_tags"),
        ):
            async with transaction() as trx:
                await remove_relations_for_entries(trx, [cataloged_entry.id])

        await assert_has_tags({cataloged_entry.id: set(), another_cataloged_entry.id: set(three_tags_ids)})

        await assert_tags_processors(entry_id=cataloged_entry.id, tag_processors={})

        await assert_tags_processors(
            entry_id=another_cataloged_entry.id,
            tag_processors={
                three_tags_ids[0]: {fake_processor_id, another_fake_processor_id},
                three_tags_ids[1]: {fake_processor_id, another_fake_processor_id},
                three_tags_ids[2]: {fake_processor_id, another_fake_processor_id},
            },
        )
